/*
    FreeRTOS V7.1.0 - Copyright (C) 2011 Real Time Engineers Ltd.

    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    ***NOTE*** The exception to the GPL is included to allow you to distribute
    a combined work that includes FreeRTOS without being obliged to provide the
    source code for proprietary components outside of the FreeRTOS kernel.
    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public 
    License and the FreeRTOS license exception along with FreeRTOS; if not it 
    can be viewed here: http://www.freertos.org/a00114.html and also obtained 
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/

/* Standard includes. */
#include <stdio.h>

/* MPLAB includes. */
#include <p32xxxx.h>
#include <plib.h>

/* FreeRTOS.org includes. */
#include "FreeRTOS.h"
#include "task.h"

/* Demo includes. */
#include "basic_io.h"

/* PIC32 configuration. */
#pragma config FPLLODIV = DIV_1, FPLLMUL = MUL_20, FPLLIDIV = DIV_2
#pragma config FWDTEN = OFF, FPBDIV = DIV_2, POSCMOD = XT, FNOSC = PRIPLL, CP = OFF
#pragma config FSRSSEL = PRIORITY_7

/* The task functions. */
static void prvDemoTask( void *pvParameters );
static void prvStatsTask( void *pvParameters );

/* The time base for the run time stats is generated by timer 2 (T2).  T2 is
configured to to generate an interrupt each time its count value overflows.  
By counting the overflows a 32-bit time value can be constructed from the 16-bit 
timer.  The overflows are counted in the prvTMR2OverflowIntHandler() ISR. The
interrupt is configured to use IPL6, which is above any priority used by the
kernel.  Therefore the kernel cannot introduce any timing jitter into the T2
interrupt.  Note that, for this example, the FreeRTOS assembly file wrappers
are not used for this interrupt - which means the interrupt will not use the
dedicated interrupt stack, and the IPL setting will be used by the compiler
generated prologue code.  The interrupt priority used is set by the call to
ConfigIntTimer2(), which must also be 6 to match the IPL setting. */
void __attribute__( (interrupt(ipl6), vector(_TIMER_2_VECTOR))) prvTMR2OverflowIntHandler( void );

/* Timer 2 (T2) is configured to generate an interrupt when it overflows.
ulTMR2OverflowCount is incremented in prvTMR2OverflowIntHandler() to count the
number of overflows that have occurred. */
volatile unsigned long ulTMR2OverflowCount = 0;

/* Basic hardware and debug interface configuration. */
void vSetupEnvironment( void );

/*-----------------------------------------------------------*/

int main( void )
{
	/* Configure both the hardware and the debug interface. */
	vSetupEnvironment();

	/* Two tasks are created above the idle task priority (at tskIDLE_PRIORITY + 
	1) that just use up some processing time before blocking for a very short 
	period. */
	xTaskCreate( prvDemoTask, ( signed char * ) "Demo1", 240, NULL,	tskIDLE_PRIORITY + 1, NULL );
	xTaskCreate( prvDemoTask, ( signed char * ) "Demo2", 240, NULL, tskIDLE_PRIORITY + 1, NULL );

	/* The highest priority task (created at tskIDLE_PRIORITY + 2 ) runs every 
	five seconds and prints out the run time stats. */
	xTaskCreate( prvStatsTask, ( signed char * ) "Stats", 240, NULL, tskIDLE_PRIORITY + 2, NULL );

	/* Start the scheduler so the tasks start executing. */
	vTaskStartScheduler();	
	
	/* If all is well we will never reach here as the scheduler will now be
	running.  If we do reach here then it is likely that there was insufficient
	heap available for the idle task to be created. */
	for( ;; );
	return 0;
}
/*-----------------------------------------------------------*/

static void prvDemoTask( void *pvParameters )
{
volatile unsigned long ulLoopCounter;
const unsigned long ulMaxLoopCount = 0x1fffUL;
portTickType xLastExecutionTime;

/* The task will run every 5 milliseconds. */
const portTickType xBlockPeriod = ( 5 / portTICK_RATE_MS );

	/* Initialise xLastExecutionTime to the current time.  This is the only
	time this variable needs to be written to explicitly.  Afterwards it is
	updated internally within the xTaskDelayUnitl() API function. */
	xLastExecutionTime = xTaskGetTickCount();

	/* As per most tasks, this task is implemented in an infinite loop. */
	for( ;; )
	{
		/* Wait until it is time to run this task again. */
		vTaskDelayUntil( &xLastExecutionTime, xBlockPeriod );
		
		/* This loop is just to ensure the task uses up enough processing time
		to register in the run time statistics. */
		for( ulLoopCounter = 0; ulLoopCounter < ulMaxLoopCount; ulLoopCounter++ )
		{
			/* There is nothing to do here.  Just perform a "no operation" to
			ensure there are some instructions generated. */
			portNOP();
		}
	}
}
/*-----------------------------------------------------------*/

static void prvStatsTask( void *pvParameters )
{
portTickType xLastExecutionTime;

/* The buffer used to hold the run time stats text needs to be quite large.  It 
is therefore declared static to ensure it is not allocated on the task stack.  
This makes this function non re-entrant. */
static signed char cStringBuffer[ 512 ]; 

/* The task will run every 5 seconds. */
const portTickType xBlockPeriod = ( 5000 / portTICK_RATE_MS );

	/* Initialise xLastExecutionTime to the current time.  This is the only
	time this variable needs to be written to explicitly.  Afterwards it is
	updated internally within the xTaskDelayUntil() API function. */
	xLastExecutionTime = xTaskGetTickCount();

	/* As per most tasks, this task is implemented in an infinite loop. */
	for( ;; )
	{
		/* Wait until it is time to run this task again. */
		vTaskDelayUntil( &xLastExecutionTime, xBlockPeriod );

		/* Generate a text table from the run time stats.  This must fit into
		the cStringBuffer array. */
		vTaskGetRunTimeStats( cStringBuffer );
		
		/* Print out column headings for the run time stats table. */
		printf( "\nTask\t\tAbs\t\t\t%%\n" );
		fflush( stdout );
		printf( "-------------------------------------------------------------\n" );
		fflush( stdout );
		
		/* Print out the run time stats themselves.  The table of data contains
		multiple lines, so the vPrintMultipleLines() function is called instead
		of calling printf() directly.  vPrintMultipleLines() simply calls
		printf() on each line individually, to ensure the line buffering words
		as expected. */
		vPrintMultipleLines( cStringBuffer );
		fflush( stdout );
	}
}
/*-----------------------------------------------------------*/

void vSetupTimerForRunTimeStats( void )
{
const unsigned int uiMax16BitValue = 0xffff;

	/* Clock T2 from PCLK/256, and interrupt on every overflow (actually, when
	the counter value reaches its maxium creating, in effect, and overflow
	interrupt. */
	OpenTimer2( ( T2_ON | T2_PS_1_256 | T2_SOURCE_INT ), uiMax16BitValue );

	/* Configure the T2 interrupt to use priority 6 - which is above any
	priority used by the kernel. */
	ConfigIntTimer2( T2_INT_ON | 6 );
}
/*-----------------------------------------------------------*/

void prvTMR2OverflowIntHandler( void )
{
	/* Just count the number of interrupts.  This count is shifted to the two most
	significant bytes and combined with the current T2 counter value to generate
	a 32-bit run time counter value from a 16-bit counter. */
	ulTMR2OverflowCount++;

	/* Clear the T2 interrupt. */
	mT2ClearIntFlag();
}	
/*-----------------------------------------------------------*/

void vSetupEnvironment( void )
{
	/* Setup the main clock for maximum performance, and the peripheral clock
	to equal the main clock divided by 2. */
	SYSTEMConfigPerformance( configCPU_CLOCK_HZ );
	mOSCSetPBDIV( OSC_PB_DIV_2 );
	
	/* Enable global interrupt handling. */
	INTEnableSystemMultiVectoredInt();

	/* Initialise the debug utils library to enable strings printed from the
	demo source code to be displayed in the MPLAB IDE. */
	DBINIT();

	/* Set the console output to line buffered mode. */
	vSetIOBuffering();
}
/*-----------------------------------------------------------*/

